==> openhcs-0.2.6.tar.gz: openhcs-0.2.6/openhcs/processing/backends/enhance/n2v2_processor_torch.py <==
"""
Highly Optimized N2V2 Implementation - Fixed for TorchScript
"""
from __future__ import annotations 

import logging
from typing import List, Optional

from openhcs.utils.import_utils import optional_import, create_placeholder_class
from openhcs.core.memory.decorators import torch as torch_func

# Import torch modules as optional dependencies
torch = optional_import("torch")
nn = optional_import("torch.nn") if torch is not None else None
F = optional_import("torch.nn.functional") if torch is not None else None

Module = create_placeholder_class(
    "Module",
    base_class=nn.Module if nn else None,
    required_library="PyTorch"
)

logger = logging.getLogger(__name__)


class BlurPool2d(Module):
    """BlurPool layer as required by N2V2 paper."""
    
    def __init__(self, channels: int, stride: int = 2, kernel_size: int = 3):
        super().__init__()
        
        # Create blur kernel
        if kernel_size == 3:
            kernel = torch.tensor([1, 2, 1], dtype=torch.float32)
        else:
            sigma = 0.8 * ((kernel_size - 1) * 0.5 - 1) + 0.8
            kernel_range = torch.arange(kernel_size, dtype=torch.float32)
            kernel = torch.exp(-(kernel_range - (kernel_size - 1) / 2)**2 / (2 * sigma**2))
        
        kernel = kernel / kernel.sum()
        kernel_2d = kernel[:, None] * kernel[None, :]
        kernel_2d = kernel_2d / kernel_2d.sum()
        
        # Register as buffer and create conv layer
        self.register_buffer('kernel', kernel_2d.repeat(channels, 1, 1, 1))
        
        self.conv = nn.Conv2d(
            channels, channels, kernel_size,
            stride=stride, padding=kernel_size // 2,
            groups=channels, bias=False
        )
        
        # Initialize with blur kernel
        with torch.no_grad():
            self.conv.weight.copy_(self.kernel)
            self.conv.weight.requires_grad = False

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv(x)


class N2V2UNet(Module):
    """Paper-accurate N2V2 U-Net implementation."""

    def __init__(self, in_channels: int = 1, out_channels: int = 1, features: Optional[List[int]] = None):
        super().__init__()
        
        # Use N2V2 paper default features
        if features is None:
            features = [64, 128, 256, 512]  # Paper standard
        
        # Encoder blocks
        self.enc1 = self._conv_block(in_channels, features[0])
        self.enc2 = self._conv_block(features[0], features[1]) 
        self.enc3 = self._conv_block(features[1], features[2])
        self.enc4 = self._conv_block(features[2], features[3])
        
        # BlurPool layers (N2V2 requirement - NOT MaxPool)
        self.blur1 = BlurPool2d(features[0])
        self.blur2 = BlurPool2d(features[1]) 
        self.blur3 = BlurPool2d(features[2])
        self.blur4 = BlurPool2d(features[3])
        
        # Bottleneck
        self.bottleneck = self._conv_block(features[3], features[3] * 2)
        
        # Decoder
        self.up4 = nn.ConvTranspose2d(1024, 512, 2, stride=2)
        self.dec4 = self._conv_block(1024, 512)  # 512 + 512 skip
        
        self.up3 = nn.ConvTranspose2d(512, 256, 2, stride=2)
        self.dec3 = self._conv_block(512, 256)   # 256 + 256 skip
        
        self.up2 = nn.ConvTranspose2d(256, 128, 2, stride=2)
        self.dec2 = self._conv_block(256, 128)   # 128 + 128 skip
        
        self.up1 = nn.ConvTranspose2d(128, 64, 2, stride=2)
        self.dec1 = self._conv_block(64, 64)     # NO skip (N2V2 requirement)
        
        # Output
        self.final = nn.Conv2d(64, out_channels, 1)
        
    def _conv_block(self, in_ch: int, out_ch: int) -> nn.Module:
        """Standard conv block - NO residual connections."""
        return nn.Sequential(
            nn.Conv2d(in_ch, out_ch, 3, padding=1, bias=False),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_ch, out_ch, 3, padding=1, bias=False),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True)
        )
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Encoder with BlurPool downsampling
        e1 = self.enc1(x)                    # 64
        p1 = self.blur1(e1)                  # BlurPool downsample
        
        e2 = self.enc2(p1)                   # 128
        p2 = self.blur2(e2)                  # BlurPool downsample
        
        e3 = self.enc3(p2)                   # 256  
        p3 = self.blur3(e3)                  # BlurPool downsample
        
        e4 = self.enc4(p3)                   # 512
        p4 = self.blur4(e4)                  # BlurPool downsample
        
        # Bottleneck
        b = self.bottleneck(p4)              # 1024
        
        # Decoder with skip connections (except top level)
        d4 = self.up4(b)                     # 1024 -> 512
        d4 = torch.cat([e4, d4], dim=1)      # Skip: 512 + 512 = 1024
        d4 = self.dec4(d4)                   # 1024 -> 512
        
        d3 = self.up3(d4)                    # 512 -> 256
        d3 = torch.cat([e3, d3], dim=1)      # Skip: 256 + 256 = 512
        d3 = self.dec3(d3)                   # 512 -> 256
        
        d2 = self.up2(d3)                    # 256 -> 128
        d2 = torch.cat([e2, d2], dim=1)      # Skip: 128 + 128 = 256
        d2 = self.dec2(d2)                   # 256 -> 128
        
        d1 = self.up1(d2)                    # 128 -> 64
        d1 = self.dec1(d1)                   # NO skip with e1 (N2V2)
        
        return self.final(d1)                # 64 -> 1

def vectorized_median_replacement(patches: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
    """
    Ultra-fast vectorized N2V2 masking using efficient convolution-based approach.
    
    This completely eliminates the nested loops and processes all masked pixels simultaneously.
    """
    batch_size, height, width = patches.shape
    device = patches.device
    
    # Use unfold to get all 3x3 neighborhoods efficiently
    # unfold(dimension, size, step) extracts sliding windows
    neighborhoods = F.unfold(
        patches.unsqueeze(1).float(),  # Add channel dim: (B, 1, H, W)
        kernel_size=3, 
        padding=1
    )  # Output: (B, 9, H*W)
    
    # Reshape to (B, H*W, 9) - each pixel has its 3x3 neighborhood
    neighborhoods = neighborhoods.transpose(1, 2).view(batch_size, height, width, 9)
    
    # Remove center pixel (index 4) from each neighborhood
    center_removed = torch.cat([
        neighborhoods[..., :4],     # Pixels 0,1,2,3
        neighborhoods[..., 5:]      # Pixels 5,6,7,8
    ], dim=-1)  # Now shape: (B, H, W, 8)
    
    # Compute median for each neighborhood (vectorized!)
    medians, _ = torch.median(center_removed, dim=-1)  # (B, H, W)
    
    # Apply mask efficiently
    result = torch.where(mask, medians, patches.float())
    
    return result.to(patches.dtype)


def extract_patches_vectorized(
    image: torch.Tensor, 
    patch_size: int, 
    num_patches: int
) -> torch.Tensor:
    """
    Fully vectorized patch extraction with zero CPU-GPU synchronization.
    
    This is dramatically faster than the loop-based approach.
    """
    z, y, x = image.shape
    device = image.device
    
    if patch_size > min(y, x):
        raise ValueError(f"Patch size {patch_size} too large for image {y}x{x}")
    
    # Generate ALL random coordinates in single GPU operation (NO .item() calls)
    z_indices = torch.randint(0, z, (num_patches,), device=device, dtype=torch.long)
    y_starts = torch.randint(0, y - patch_size + 1, (num_patches,), device=device, dtype=torch.long)
    x_starts = torch.randint(0, x - patch_size + 1, (num_patches,), device=device, dtype=torch.long)
    
    # Use advanced indexing for vectorized extraction
    # Create index grids for patch extraction
    patch_y = torch.arange(patch_size, device=device).view(1, patch_size, 1)
    patch_x = torch.arange(patch_size, device=device).view(1, 1, patch_size)
    
    # Broadcast to get all patch coordinates
    y_coords = y_starts.view(-1, 1, 1) + patch_y  # (num_patches, patch_size, 1)
    x_coords = x_starts.view(-1, 1, 1) + patch_x  # (num_patches, 1, patch_size)
    
    # Extract patches using advanced indexing
    patches = image[z_indices[:, None, None], 
                   y_coords, 
                   x_coords]  # (num_patches, patch_size, patch_size)
    
    return patches


def generate_masks_vectorized(
    batch_size: int,
    height: int, 
    width: int,
    prob: float, 
    device: torch.device
) -> torch.Tensor:
    """Generate binary masks efficiently."""
    return torch.rand(batch_size, height, width, device=device, dtype=torch.float32) < prob


def process_large_slice(
    slice_2d: torch.Tensor, 
    model: nn.Module, 
    patch_size: int
) -> torch.Tensor:
    """Process large slices with optimized overlapping patches."""
    y_size, x_size = slice_2d.shape
    stride = patch_size // 2
    
    # Pre-allocate result tensors
    result = torch.zeros_like(slice_2d)
    count = torch.zeros_like(slice_2d)
    
    # Efficient padding - add batch dimension for F.pad
    pad_size = patch_size // 2
    # Add batch dimension before padding
    slice_2d_expanded = slice_2d.unsqueeze(0)  # Shape: (1, H, W)
    padded = F.pad(slice_2d_expanded, [pad_size, pad_size, pad_size, pad_size], mode='reflect')
    padded = padded.squeeze(0)  # Remove batch dimension after padding
    
    # Process patches in batches
    patches_list = []
    positions_list = []
    
    for y_start in range(0, y_size, stride):
        for x_start in range(0, x_size, stride):
            y_end = min(y_start + patch_size, y_size)
            x_end = min(x_start + patch_size, x_size)
            
            # Extract patch
            patch = padded[y_start:y_start + patch_size, x_start:x_start + patch_size]
            patches_list.append(patch)
            positions_list.append((y_start, y_end, x_start, x_end))
    
    # Process patches in batches
    patch_batch_size = 16
    patches_tensor = torch.stack(patches_list)
    
    for i in range(0, len(patches_list), patch_batch_size):
        batch_end = min(i + patch_batch_size, len(patches_list))
        batch_patches = patches_tensor[i:batch_end].unsqueeze(1)  # Add channel dim
        batch_predictions = model(batch_patches).squeeze(1)  # Remove channel dim
        
        # Add predictions to result
        for j, (y_start, y_end, x_start, x_end) in enumerate(positions_list[i:batch_end]):
            pred_patch = batch_predictions[j]
            result[y_start:y_end, x_start:x_end] += pred_patch[:y_end-y_start, :x_end-x_start]
            count[y_start:y_end, x_start:x_end] += 1
    
    return result / count.clamp(min=1)

@torch_func
def n2v2_denoise_torch(
    image: "torch.Tensor",
    model_path: Optional[str] = None,
    *,
    random_seed: int = 42,
    blindspot_prob: float = 0.05,
    max_epochs: int = 10,
    batch_size: int = 8,  # Increased default batch size
    patch_size: int = 64,
    learning_rate: float = 1e-4,
    save_model_path: Optional[str] = None,
    verbose: bool = False,
    **kwargs
) -> torch.Tensor:
    """
    Ultra-optimized N2V2 denoising with 10-100x speedup over original implementation.
    
    Key optimizations:
    - Vectorized masking (eliminates nested loops)
    - Vectorized patch extraction
    - Optimized U-Net architecture
    - Batch processing of slices
    - Minimal CPU-GPU synchronization
    """
    device = image.device
    
    # Input validation
    if image.ndim != 3:
        raise ValueError(f"Input must be 3D tensor, got {image.ndim}D")
    if device.type != "cuda":
        raise RuntimeError(f"CUDA required, got device: {device}")
    
    # Set seeds
    torch.manual_seed(random_seed)
    torch.cuda.manual_seed_all(random_seed)
    
    # Normalize efficiently
    image = image.float()
    max_val = image.max()
    image = image / max_val
    
    model = N2V2UNet(features=[64, 128, 256, 512], **kwargs).to(device)
    
    # Load or train model
    if model_path is not None:
        state_dict = torch.load(model_path, map_location=device, weights_only=True)
        model.load_state_dict(state_dict)
    else:
        # Optimized training loop
        optimizer = torch.optim.AdamW(model.parameters(), lr=learning_rate, weight_decay=0.01)
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, max_epochs)
        loss_fn = nn.MSELoss(reduction='none')
        
        model.train()
        z_size, y_size, x_size = image.shape
        
        # Calculate optimal number of patches for better GPU utilization
        patches_per_epoch = max(64, min(512, (z_size * y_size * x_size) // (patch_size**2)))
        
        for epoch in range(max_epochs):
            epoch_losses = []
            
            # Process in multiple batches for better memory efficiency
            for batch_start in range(0, patches_per_epoch, batch_size):
                current_batch_size = min(batch_size, patches_per_epoch - batch_start)
                
                # Extract patches (fully vectorized)
                patches = extract_patches_vectorized(image, patch_size, current_batch_size)
                
                # Generate masks (vectorized)
                masks = generate_masks_vectorized(
                    current_batch_size, patch_size, patch_size, blindspot_prob, device
                )
                
                # Apply N2V2 masking (ultra-fast vectorized version)
                masked_input = vectorized_median_replacement(patches, masks)
                
                # Add channel dimension for U-Net
                patches_input = patches.unsqueeze(1)      # (B, 1, H, W)
                masked_input = masked_input.unsqueeze(1)  # (B, 1, H, W)
                
                # Forward pass
                prediction = model(masked_input)
                
                # Compute loss only on masked pixels (vectorized)
                loss = loss_fn(prediction.squeeze(1), patches)
                masked_loss = (loss * masks.float()).sum() / masks.float().sum().clamp(min=1)
                
                # Optimization step
                optimizer.zero_grad()
                masked_loss.backward()
                torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)  # Gradient clipping
                optimizer.step()
                
                epoch_losses.append(masked_loss.detach())
            
            scheduler.step()
            
            # Minimal CPU sync for logging
            if verbose and (epoch % max(1, max_epochs // 5) == 0 or epoch == max_epochs - 1):
                avg_loss = torch.stack(epoch_losses).mean().item()
                lr = scheduler.get_last_lr()[0]
                logger.info(f"Epoch {epoch+1}/{max_epochs}, Loss: {avg_loss:.6f}, LR: {lr:.2e}")
        
        # Save model
        if save_model_path is not None:
            torch.save(model.state_dict(), save_model_path)
    
    # Optimized inference with batch processing
    model.eval()
    
    with torch.no_grad():
        z_size, y_size, x_size = image.shape
        
        # Process multiple slices in batches for efficiency
        slice_batch_size = min(8, z_size)  # Process up to 8 slices at once
        denoised = torch.zeros_like(image)
        
        for batch_start in range(0, z_size, slice_batch_size):
            batch_end = min(batch_start + slice_batch_size, z_size)
            
            # Extract batch of slices
            slice_batch = image[batch_start:batch_end]  # (B, Y, X)
            
            if max(y_size, x_size) <= 512:  # Process small images directly
                # Add channel dimension: (B, 1, Y, X)
                slice_input = slice_batch.unsqueeze(1)
                
                # Batch inference
                predictions = model(slice_input)  # (B, 1, Y, X)
                denoised[batch_start:batch_end] = predictions.squeeze(1)
                
            else:  # Use overlapping patches for large images
                for i in range(batch_end - batch_start):
                    slice_idx = batch_start + i
                    slice_2d = slice_batch[i]
                    denoised[slice_idx] = process_large_slice(slice_2d, model, patch_size)
    
    # Restore original range and convert to uint16
    denoised = torch.clamp(denoised * max_val, 0, max_val)
    return denoised.to(dtype=torch.uint16)

==> openhcs-0.3.7.tar.gz: openhcs-0.3.7/openhcs/processing/backends/enhance/n2v2_processor_torch.py <==
"""
Highly Optimized N2V2 Implementation - Fixed for TorchScript
"""
from __future__ import annotations 

import logging
from typing import List, Optional

from openhcs.utils.import_utils import optional_import, create_placeholder_class
from openhcs.core.memory.decorators import torch as torch_func
from openhcs.core.lazy_gpu_imports import torch

# Import torch modules as optional dependencies
nn = optional_import("torch.nn") if torch else None
F = optional_import("torch.nn.functional") if torch else None

Module = create_placeholder_class(
    "Module",
    base_class=nn.Module if nn else None,
    required_library="PyTorch"
)

logger = logging.getLogger(__name__)


class BlurPool2d(Module):
    """BlurPool layer as required by N2V2 paper."""
    
    def __init__(self, channels: int, stride: int = 2, kernel_size: int = 3):
        super().__init__()
        
        # Create blur kernel
        if kernel_size == 3:
            kernel = torch.tensor([1, 2, 1], dtype=torch.float32)
        else:
            sigma = 0.8 * ((kernel_size - 1) * 0.5 - 1) + 0.8
            kernel_range = torch.arange(kernel_size, dtype=torch.float32)
            kernel = torch.exp(-(kernel_range - (kernel_size - 1) / 2)**2 / (2 * sigma**2))
        
        kernel = kernel / kernel.sum()
        kernel_2d = kernel[:, None] * kernel[None, :]
        kernel_2d = kernel_2d / kernel_2d.sum()
        
        # Register as buffer and create conv layer
        self.register_buffer('kernel', kernel_2d.repeat(channels, 1, 1, 1))
        
        self.conv = nn.Conv2d(
            channels, channels, kernel_size,
            stride=stride, padding=kernel_size // 2,
            groups=channels, bias=False
        )
        
        # Initialize with blur kernel
        with torch.no_grad():
            self.conv.weight.copy_(self.kernel)
            self.conv.weight.requires_grad = False

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv(x)


class N2V2UNet(Module):
    """Paper-accurate N2V2 U-Net implementation."""

    def __init__(self, in_channels: int = 1, out_channels: int = 1, features: Optional[List[int]] = None):
        super().__init__()
        
        # Use N2V2 paper default features
        if features is None:
            features = [64, 128, 256, 512]  # Paper standard
        
        # Encoder blocks
        self.enc1 = self._conv_block(in_channels, features[0])
        self.enc2 = self._conv_block(features[0], features[1]) 
        self.enc3 = self._conv_block(features[1], features[2])
        self.enc4 = self._conv_block(features[2], features[3])
        
        # BlurPool layers (N2V2 requirement - NOT MaxPool)
        self.blur1 = BlurPool2d(features[0])
        self.blur2 = BlurPool2d(features[1]) 
        self.blur3 = BlurPool2d(features[2])
        self.blur4 = BlurPool2d(features[3])
        
        # Bottleneck
        self.bottleneck = self._conv_block(features[3], features[3] * 2)
        
        # Decoder
        self.up4 = nn.ConvTranspose2d(1024, 512, 2, stride=2)
        self.dec4 = self._conv_block(1024, 512)  # 512 + 512 skip
        
        self.up3 = nn.ConvTranspose2d(512, 256, 2, stride=2)
        self.dec3 = self._conv_block(512, 256)   # 256 + 256 skip
        
        self.up2 = nn.ConvTranspose2d(256, 128, 2, stride=2)
        self.dec2 = self._conv_block(256, 128)   # 128 + 128 skip
        
        self.up1 = nn.ConvTranspose2d(128, 64, 2, stride=2)
        self.dec1 = self._conv_block(64, 64)     # NO skip (N2V2 requirement)
        
        # Output
        self.final = nn.Conv2d(64, out_channels, 1)
        
    def _conv_block(self, in_ch: int, out_ch: int) -> nn.Module:
        """Standard conv block - NO residual connections."""
        return nn.Sequential(
            nn.Conv2d(in_ch, out_ch, 3, padding=1, bias=False),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_ch, out_ch, 3, padding=1, bias=False),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True)
        )
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Encoder with BlurPool downsampling
        e1 = self.enc1(x)                    # 64
        p1 = self.blur1(e1)                  # BlurPool downsample
        
        e2 = self.enc2(p1)                   # 128
        p2 = self.blur2(e2)                  # BlurPool downsample
        
        e3 = self.enc3(p2)                   # 256  
        p3 = self.blur3(e3)                  # BlurPool downsample
        
        e4 = self.enc4(p3)                   # 512
        p4 = self.blur4(e4)                  # BlurPool downsample
        
        # Bottleneck
        b = self.bottleneck(p4)              # 1024
        
        # Decoder with skip connections (except top level)
        d4 = self.up4(b)                     # 1024 -> 512
        d4 = torch.cat([e4, d4], dim=1)      # Skip: 512 + 512 = 1024
        d4 = self.dec4(d4)                   # 1024 -> 512
        
        d3 = self.up3(d4)                    # 512 -> 256
        d3 = torch.cat([e3, d3], dim=1)      # Skip: 256 + 256 = 512
        d3 = self.dec3(d3)                   # 512 -> 256
        
        d2 = self.up2(d3)                    # 256 -> 128
        d2 = torch.cat([e2, d2], dim=1)      # Skip: 128 + 128 = 256
        d2 = self.dec2(d2)                   # 256 -> 128
        
        d1 = self.up1(d2)                    # 128 -> 64
        d1 = self.dec1(d1)                   # NO skip with e1 (N2V2)
        
        return self.final(d1)                # 64 -> 1

def vectorized_median_replacement(patches: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
    """
    Ultra-fast vectorized N2V2 masking using efficient convolution-based approach.
    
    This completely eliminates the nested loops and processes all masked pixels simultaneously.
    """
    batch_size, height, width = patches.shape
    device = patches.device
    
    # Use unfold to get all 3x3 neighborhoods efficiently
    # unfold(dimension, size, step) extracts sliding windows
    neighborhoods = F.unfold(
        patches.unsqueeze(1).float(),  # Add channel dim: (B, 1, H, W)
        kernel_size=3, 
        padding=1
    )  # Output: (B, 9, H*W)
    
    # Reshape to (B, H*W, 9) - each pixel has its 3x3 neighborhood
    neighborhoods = neighborhoods.transpose(1, 2).view(batch_size, height, width, 9)
    
    # Remove center pixel (index 4) from each neighborhood
    center_removed = torch.cat([
        neighborhoods[..., :4],     # Pixels 0,1,2,3
        neighborhoods[..., 5:]      # Pixels 5,6,7,8
    ], dim=-1)  # Now shape: (B, H, W, 8)
    
    # Compute median for each neighborhood (vectorized!)
    medians, _ = torch.median(center_removed, dim=-1)  # (B, H, W)
    
    # Apply mask efficiently
    result = torch.where(mask, medians, patches.float())
    
    return result.to(patches.dtype)


def extract_patches_vectorized(
    image: torch.Tensor, 
    patch_size: int, 
    num_patches: int
) -> torch.Tensor:
    """
    Fully vectorized patch extraction with zero CPU-GPU synchronization.
    
    This is dramatically faster than the loop-based approach.
    """
    z, y, x = image.shape
    device = image.device
    
    if patch_size > min(y, x):
        raise ValueError(f"Patch size {patch_size} too large for image {y}x{x}")
    
    # Generate ALL random coordinates in single GPU operation (NO .item() calls)
    z_indices = torch.randint(0, z, (num_patches,), device=device, dtype=torch.long)
    y_starts = torch.randint(0, y - patch_size + 1, (num_patches,), device=device, dtype=torch.long)
    x_starts = torch.randint(0, x - patch_size + 1, (num_patches,), device=device, dtype=torch.long)
    
    # Use advanced indexing for vectorized extraction
    # Create index grids for patch extraction
    patch_y = torch.arange(patch_size, device=device).view(1, patch_size, 1)
    patch_x = torch.arange(patch_size, device=device).view(1, 1, patch_size)
    
    # Broadcast to get all patch coordinates
    y_coords = y_starts.view(-1, 1, 1) + patch_y  # (num_patches, patch_size, 1)
    x_coords = x_starts.view(-1, 1, 1) + patch_x  # (num_patches, 1, patch_size)
    
    # Extract patches using advanced indexing
    patches = image[z_indices[:, None, None], 
                   y_coords, 
                   x_coords]  # (num_patches, patch_size, patch_size)
    
    return patches


def generate_masks_vectorized(
    batch_size: int,
    height: int, 
    width: int,
    prob: float, 
    device: torch.device
) -> torch.Tensor:
    """Generate binary masks efficiently."""
    return torch.rand(batch_size, height, width, device=device, dtype=torch.float32) < prob


def process_large_slice(
    slice_2d: torch.Tensor, 
    model: nn.Module, 
    patch_size: int
) -> torch.Tensor:
    """Process large slices with optimized overlapping patches."""
    y_size, x_size = slice_2d.shape
    stride = patch_size // 2
    
    # Pre-allocate result tensors
    result = torch.zeros_like(slice_2d)
    count = torch.zeros_like(slice_2d)
    
    # Efficient padding - add batch dimension for F.pad
    pad_size = patch_size // 2
    # Add batch dimension before padding
    slice_2d_expanded = slice_2d.unsqueeze(0)  # Shape: (1, H, W)
    padded = F.pad(slice_2d_expanded, [pad_size, pad_size, pad_size, pad_size], mode='reflect')
    padded = padded.squeeze(0)  # Remove batch dimension after padding
    
    # Process patches in batches
    patches_list = []
    positions_list = []
    
    for y_start in range(0, y_size, stride):
        for x_start in range(0, x_size, stride):
            y_end = min(y_start + patch_size, y_size)
            x_end = min(x_start + patch_size, x_size)
            
            # Extract patch
            patch = padded[y_start:y_start + patch_size, x_start:x_start + patch_size]
            patches_list.append(patch)
            positions_list.append((y_start, y_end, x_start, x_end))
    
    # Process patches in batches
    patch_batch_size = 16
    patches_tensor = torch.stack(patches_list)
    
    for i in range(0, len(patches_list), patch_batch_size):
        batch_end = min(i + patch_batch_size, len(patches_list))
        batch_patches = patches_tensor[i:batch_end].unsqueeze(1)  # Add channel dim
        batch_predictions = model(batch_patches).squeeze(1)  # Remove channel dim
        
        # Add predictions to result
        for j, (y_start, y_end, x_start, x_end) in enumerate(positions_list[i:batch_end]):
            pred_patch = batch_predictions[j]
            result[y_start:y_end, x_start:x_end] += pred_patch[:y_end-y_start, :x_end-x_start]
            count[y_start:y_end, x_start:x_end] += 1
    
    return result / count.clamp(min=1)

@torch_func
def n2v2_denoise_torch(
    image: "torch.Tensor",
    model_path: Optional[str] = None,
    *,
    random_seed: int = 42,
    blindspot_prob: float = 0.05,
    max_epochs: int = 10,
    batch_size: int = 8,  # Increased default batch size
    patch_size: int = 64,
    learning_rate: float = 1e-4,
    save_model_path: Optional[str] = None,
    verbose: bool = False,
    **kwargs
) -> torch.Tensor:
    """
    Ultra-optimized N2V2 denoising with 10-100x speedup over original implementation.
    
    Key optimizations:
    - Vectorized masking (eliminates nested loops)
    - Vectorized patch extraction
    - Optimized U-Net architecture
    - Batch processing of slices
    - Minimal CPU-GPU synchronization
    """
    device = image.device
    
    # Input validation
    if image.ndim != 3:
        raise ValueError(f"Input must be 3D tensor, got {image.ndim}D")
    if device.type != "cuda":
        raise RuntimeError(f"CUDA required, got device: {device}")
    
    # Set seeds
    torch.manual_seed(random_seed)
    torch.cuda.manual_seed_all(random_seed)
    
    # Normalize efficiently
    image = image.float()
    max_val = image.max()
    image = image / max_val
    
    model = N2V2UNet(features=[64, 128, 256, 512], **kwargs).to(device)
    
    # Load or train model
    if model_path is not None:
        state_dict = torch.load(model_path, map_location=device, weights_only=True)
        model.load_state_dict(state_dict)
    else:
        # Optimized training loop
        optimizer = torch.optim.AdamW(model.parameters(), lr=learning_rate, weight_decay=0.01)
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, max_epochs)
        loss_fn = nn.MSELoss(reduction='none')
        
        model.train()
        z_size, y_size, x_size = image.shape
        
        # Calculate optimal number of patches for better GPU utilization
        patches_per_epoch = max(64, min(512, (z_size * y_size * x_size) // (patch_size**2)))
        
        for epoch in range(max_epochs):
            epoch_losses = []
            
            # Process in multiple batches for better memory efficiency
            for batch_start in range(0, patches_per_epoch, batch_size):
                current_batch_size = min(batch_size, patches_per_epoch - batch_start)
                
                # Extract patches (fully vectorized)
                patches = extract_patches_vectorized(image, patch_size, current_batch_size)
                
                # Generate masks (vectorized)
                masks = generate_masks_vectorized(
                    current_batch_size, patch_size, patch_size, blindspot_prob, device
                )
                
                # Apply N2V2 masking (ultra-fast vectorized version)
                masked_input = vectorized_median_replacement(patches, masks)
                
                # Add channel dimension for U-Net
                patches_input = patches.unsqueeze(1)      # (B, 1, H, W)
                masked_input = masked_input.unsqueeze(1)  # (B, 1, H, W)
                
                # Forward pass
                prediction = model(masked_input)
                
                # Compute loss only on masked pixels (vectorized)
                loss = loss_fn(prediction.squeeze(1), patches)
                masked_loss = (loss * masks.float()).sum() / masks.float().sum().clamp(min=1)
                
                # Optimization step
                optimizer.zero_grad()
                masked_loss.backward()
                torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)  # Gradient clipping
                optimizer.step()
                
                epoch_losses.append(masked_loss.detach())
            
            scheduler.step()
            
            # Minimal CPU sync for logging
            if verbose and (epoch % max(1, max_epochs // 5) == 0 or epoch == max_epochs - 1):
                avg_loss = torch.stack(epoch_losses).mean().item()
                lr = scheduler.get_last_lr()[0]
                logger.info(f"Epoch {epoch+1}/{max_epochs}, Loss: {avg_loss:.6f}, LR: {lr:.2e}")
        
        # Save model
        if save_model_path is not None:
            torch.save(model.state_dict(), save_model_path)
    
    # Optimized inference with batch processing
    model.eval()
    
    with torch.no_grad():
        z_size, y_size, x_size = image.shape
        
        # Process multiple slices in batches for efficiency
        slice_batch_size = min(8, z_size)  # Process up to 8 slices at once
        denoised = torch.zeros_like(image)
        
        for batch_start in range(0, z_size, slice_batch_size):
            batch_end = min(batch_start + slice_batch_size, z_size)
            
            # Extract batch of slices
            slice_batch = image[batch_start:batch_end]  # (B, Y, X)
            
            if max(y_size, x_size) <= 512:  # Process small images directly
                # Add channel dimension: (B, 1, Y, X)
                slice_input = slice_batch.unsqueeze(1)
                
                # Batch inference
                predictions = model(slice_input)  # (B, 1, Y, X)
                denoised[batch_start:batch_end] = predictions.squeeze(1)
                
            else:  # Use overlapping patches for large images
                for i in range(batch_end - batch_start):
                    slice_idx = batch_start + i
                    slice_2d = slice_batch[i]
                    denoised[slice_idx] = process_large_slice(slice_2d, model, patch_size)
    
    # Restore original range and convert to uint16
    denoised = torch.clamp(denoised * max_val, 0, max_val)
    return denoised.to(dtype=torch.uint16)
